cdef cypclass AheadOfTimeChunk

DEF CHUNK_SIZE = 2
cdef AheadOfTimeChunk chunk[CHUNK_SIZE]
cdef unsigned int chunk_started = 0
# chunk_allocated[index] tells us the state of memory pointed by chunk[index]
cdef int chunk_allocated[CHUNK_SIZE]

cdef AheadOfTimeChunk chunkclass_singleton
cdef bint chunkclass_singleton_allocated = False

cdef cypclass AheadOfTimeChunk:
    """
    A stupid chunk allocator:
    Whenever the user requests a new object, fills each empty cell of
    the chunk array with fresh memory. Give to the user the first
    available chunk cell with allocated but unused memory.
    This class is reaaally stupid because it still frees memory when
    the user don't need the object anymore, so the __alloc__ will be filling
    again and again the cells we just deallocated with fresh memory.
    """
    int cell_index

    __dealloc__(self):
        global chunkclass_singleton_allocated
        global chunkclass_singleton
        global chunk_allocated
        idx = self.cell_index
        if idx >= 0:
            with gil:
                print("Object destruction : mark cell", idx, "as empty")
            chunk_allocated[idx] = -1
        else:
            with gil:
                print("Object destruction: out-of-chunk deallocation")
        if self is chunkclass_singleton:
            with gil:
                print("  Object destruction : Mark singleton flag as unallocated")
            chunkclass_singleton_allocated = False
            chunkclass_singleton = <AheadOfTimeChunk> NULL

    AheadOfTimeChunk __new__(alloc, bint singleton = False, bint must_fit_in_chunk = False):
       global chunkclass_singleton_allocated
       global chunkclass_singleton
       if singleton and chunkclass_singleton_allocated:
           with gil:
               print("Object creation : Returning already allocated singleton")
           obj = chunkclass_singleton
       else:
           obj = alloc()
           if obj is NULL and not must_fit_in_chunk:
               with gil:
                   print("Object creation : Allocating object outside of chunk")
               obj = new AheadOfTimeChunk()
               obj.cell_index = -1
           if obj is not NULL and singleton:
                with gil:
                    print("Object creation : Singleton allocated")
                chunkclass_singleton_allocated = True
                chunkclass_singleton = obj
                # We should decref here (assignment to chunkclass_singleton has incref'ed obj, we do not want that)
                del chunkclass_singleton
       if obj is NULL:
           # This is done to show something for the demo
           with gil:
               print("Object creation : Obj is NULL !")
       return obj
        
    AheadOfTimeChunk __alloc__():
        """
        Convention for chunk_allocated is:
        1 - Memory in use
        0 - Memory allocated and available
        -1 - No memory allocated
        """
        global chunk_started
        global chunk_allocated
        global chunk
        if not chunk_started:
            for i in range(CHUNK_SIZE):
                chunk_allocated[i] = -1
                chunk_started = 1
        cdef AheadOfTimeChunk ptr = <AheadOfTimeChunk> NULL
        for i in range(CHUNK_SIZE):
            if chunk_allocated[i] == -1:
                with gil:
                    print("cell", i, "is empty: allocate it")
                chunk[i] = new AheadOfTimeChunk()
                chunk[i].cell_index = i
                chunk_allocated[i] = 0
            if ptr is NULL and chunk_allocated[i] == 0:
                with gil:
                    print("cell", i, "is allocated but unused: take it")
                chunk_allocated[i] = 1
                ptr = chunk[i]
                # We must decref ptr here, because it is the same reference
                del ptr
        return ptr

cpdef void testChunk():
    print("=== Creating a basic object ===")
    basic_obj = AheadOfTimeChunk()
    print()
    print("=== Getting singleton (no previous references) ===")
    singleton_obj1 = AheadOfTimeChunk(True)
    print()
    print("=== Getting singleton (one existing reference) ===")
    singleton_obj2 = AheadOfTimeChunk(True)

    if singleton_obj1 is singleton_obj2 and singleton_obj1 is chunkclass_singleton:
        print("    Both singleton pointers are correct")
    print()
    print("=== Deleting one singleton reference (one ref left) ===")
    #del singleton_obj1
    singleton_obj1 = <AheadOfTimeChunk> NULL

    print()
    print("=== Getting again singleton (one existing reference) ===")
    singleton_obj3 = AheadOfTimeChunk(True)
    if singleton_obj3 is singleton_obj2:
        print("    Freeing a singleton instance doesn't crash things")

    print()
    print("=== Allocating an object outside of chunk ===")
    outside_allocation = AheadOfTimeChunk()
    print()
    print("=== Make a fail allocation ===")
    failed_outside_allocation = AheadOfTimeChunk(False, True)

    # Make some place in the chunk to see reuse
    print()
    print("=== Deleting one singleton reference (one ref left) ===")
    #del singleton_obj3
    singleton_obj3 = <AheadOfTimeChunk> NULL
    print()
    print("=== Deleting one singleton reference (zero refs left) ===")
    #del singleton_obj2
    singleton_obj2 = <AheadOfTimeChunk> NULL

    print()
    print("=== Allocating again a basic object (inside the chunk) ===")
    inside_chunk = AheadOfTimeChunk()
    # We can afford a very specific test here
    if inside_chunk.cell_index == 1 and inside_chunk is chunk[1]:
        print("    Chunk cell reuse went fine")

    print()
    print("=== Allocating singleton outside of chunk (no previous references) ===")
    outside_singleton = AheadOfTimeChunk(True)

    print()
    print("Remaining objects at the end of this function are:")
    print("1 The first object we created (living at cell 0)")
    print("2 The first out-of-chunk allocated object")
    print("3 The last successfully non-singleton allocated object (living at cell 1)")
    print("4 One singleton instance (living outside of chunk)")
    print("The object destruction messages should reflect this order")
    print()
